#include <stdio.h>
#include <unistd.h>
#include <time.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <libgen.h>
#include <getopt.h>
#include <poll.h>
#include "wstationd.h"


// As of the writing of this daemon, SBD maximum message length
// is hardware limited to 1960 bytes, however, the SBD format
// has a maximum size of 64k per burst. We add three bytes
// for the SBD header itself. This value largely
// dictates the memory cost per connection.
#define BUFFERSIZE 65539

// Hard limit on number of concurrent connections
#define SLOTLIMIT 128

// Timeout for each connection, in seconds
#define TIMEOUT 30


static char* exec_name;

struct connection 
{
	// Buffer to store the data stream before the dump to file
	unsigned char buff[BUFFERSIZE];
	ssize_t len; // How many bytes have been written to the buffer
	struct timeval tv; // Time connection was made
	struct in_addr addr; // Address connection was made from
};


static int write_connection(struct connection *);

static void print_help()
{
	fprintf(stdout, "usage: %s [-p port] [-b addr] directory\n", exec_name);
	fprintf(stdout, "  -p, --port <port>     Port to listen on (default: 10800)\n");
	fprintf(stdout, "  -b, --bind <address>  Address to bind to (default: 0.0.0.0)\n");
	fprintf(stdout, "  -h, --help            Print this message and exit\n");
	fprintf(stdout, "  -v, --version         Print version and exit\n");
}


int main(int argc, char* argv[])
{
	const struct option longopts[] = {
		{ "port", required_argument, NULL, 'p' },
		{ "bind", required_argument, NULL, 'b' },
		{ "version", no_argument, NULL, 'v' },
		{ "help", no_argument, NULL, 'h' },
		{ NULL, 0, NULL, 0 }
	};

	int flags = 0;
	int sockfd = 0;
	int port = 10800; // Default port
	uint32_t address = INADDR_ANY; // Default listen address

	struct connection* conns[SLOTLIMIT];
	memset(conns, 0, sizeof(struct connection*) * SLOTLIMIT);
	struct pollfd fds[SLOTLIMIT+1];
	memset(&fds, 0, sizeof(struct pollfd) * (SLOTLIMIT + 1));

	exec_name = basename(argv[0]);

	for(int ch; (ch = getopt_long(argc, argv, "l:b:p:vh0", longopts, NULL)) != -1;){
		switch(ch){
			case 'p':
				port = (int)strtol(optarg, NULL, 10);
				if(port < 1 || port > 65535){
					fprintf(stderr, "%s: invalid port number\n", exec_name);
					print_help();
					goto shutdown_error;
				}
			break;

			case 'b':
				if(inet_pton(AF_INET, optarg, &address) == -1){
					fprintf(stderr, "%s: invalid bind address\n", exec_name);
					print_help();
					goto shutdown_error;
				}
			break;

			case 'v':
				fprintf(stdout, "wstationd v%s (%s %s)\n", VERSION,__DATE__, __TIME__);
			goto shutdown_clean;

			case '?':
			case 'h':
				print_help();
			goto shutdown_clean;
		}
	}

	// Accept final argument, the directory
	if(argc == (optind+1)){
		// Fail if we can't chdir to that directory
		if(chdir(argv[argc-1]) == -1){
			fprintf(stderr, "%s: cannot change directory - %s\n",
				exec_name, strerror(errno)
			);
			goto shutdown_error;
		}
	} else {
		print_help();
		goto shutdown_error;
	}
	

	// Establish the socket
	sockfd = socket(AF_INET, SOCK_STREAM, 0);
	if(sockfd == -1){
		fprintf(stderr, "%s: cannot open socket - %s\n",
			exec_name, strerror(errno)
		);
		goto shutdown_error;
	}

	// This sets REUSE on the socket so it's easily reallocated if
	// this program dies
	flags = 1;
	setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, &flags, sizeof(flags));

	struct sockaddr_in sock;
	memset(&sock, 0, sizeof(sock));
	sock.sin_family = AF_INET;
	sock.sin_addr.s_addr = address;
	sock.sin_port = htons(port);

	// Bind our socket to the specified address and port
	if(bind(sockfd, (struct sockaddr*)&sock, sizeof(sock)) < 0){
		fprintf(stderr, "%s: cannot bind - %s\n",
			exec_name, strerror(errno)
		);
		goto shutdown_error;
	}

	// Listen on port
	if(listen(sockfd, 10) != 0){
		fprintf(stderr, "%s: cannot listen - %s\n",
			exec_name, strerror(errno)
		);
		goto shutdown_error;
	}

	// Add sockfd
	fds[0].fd = sockfd;
	fds[0].events = POLLIN;
	fds[0].revents = 0;

	while(1){
		int r = poll(fds, (SLOTLIMIT + 1), 1000);

		if(r == -1){
			fprintf(stderr, "%s: socket poll error - %s\n",
				exec_name, strerror(errno)
			);
			goto shutdown_error;
		}

		// Handle existing connections
		int slotfirst = 0, slotfree = 0;
		for(int i = 1; i <= SLOTLIMIT; i++){
			// Handle event
			if(fds[i].revents & POLLIN){
				ssize_t sz = recv(
					fds[i].fd, conns[i-1]->buff,
					(BUFFERSIZE - conns[i-1]->len), 0
				);

				if(sz > 0) conns[i-1]->len += sz;
				else {
					if(conns[i-1]->len > 3){
						uint16_t sbd_sz = *(uint16_t*)(&conns[i-1]->buff[1]);
						// Convert and add three bytes for the SBD header
						ssize_t expected = (ssize_t)htons(sbd_sz) + 3;

						// Iridium SBD format check. Is the protocol byte in place?
						// Is the reported size the same as the actual size written
						// in bytes? If so, write it out.
						if(expected == conns[i-1]->len && conns[i-1]->buff[0] == 1){
							write_connection(conns[i - 1]);
						}
					}

					free(conns[i - 1]);
					conns[i - 1] = NULL;

					close(fds[i].fd);
					fds[i].fd = 0;
					fds[i].events = 0;
				}

			// Sweep up dead connections
			} else if(fds[i].fd){
				struct timeval tv;
				gettimeofday(&tv, NULL);

				if((tv.tv_sec - conns[i - 1]->tv.tv_sec) >= TIMEOUT){
					free(conns[i - 1]);
					conns[i - 1] = NULL;

					close(fds[i].fd);
					fds[i].fd = fds[i].events = 0;
				}
			}

			// Count total free slots
			if(!fds[i].fd){
				++slotfree;
				// Note the first available slot
				if(!slotfirst) slotfirst = i;
			}

			// Clear events
			fds[i].revents = 0;
		}

		// Handle new connections
		if((fds[0].revents & POLLIN) && slotfree){
			struct sockaddr_in client;
			socklen_t client_sz = sizeof(client);

			int fd = accept(sockfd, (struct sockaddr*)&client, &client_sz);
			if(fd == -1){
				fprintf(stderr, "%s: accept failed - %s\n",
					exec_name, strerror(errno)
				);
				continue;
			}
			
			// Get current flags before setting nonblock
			if((flags = fcntl(fd, F_GETFL)) == -1){
				fprintf(stderr, "%s: cannot get connection flags - %s\n",
					exec_name, strerror(errno)
				);
				close(fd);
				continue;
			}

			// Set nonblocking on socket
			if(fcntl(fd, F_SETFL, flags | O_NONBLOCK) == -1){
				fprintf(stderr, "%s: cannot set connection flags - %s\n",
					exec_name, strerror(errno)
				);
				close(fd);
				continue;
			}

			// Allocate memory for the connection's state. 
			conns[slotfirst - 1] = malloc(sizeof(struct connection));
			if(conns[slotfirst - 1] == NULL){
				fprintf(stderr, "%s: out of memory\n", exec_name);
				close(fd);
				continue;
			}

			conns[slotfirst - 1]->len = 0; // reset buffer
			conns[slotfirst - 1]->addr = client.sin_addr; // save source ip
			gettimeofday(&conns[slotfirst - 1]->tv, NULL); // mark connection time

			// Add to poll list
			fds[slotfirst].fd = fd;
			fds[slotfirst].events = POLLIN;
			fds[slotfirst].revents = 0;
			slotfree--;

			fds[0].revents = 0;
		}

		// Reset polling event on listener socket - 
		// if there's no free slots, stop polling the listener
		if(slotfree) fds[0].events = POLLIN;
		else fds[0].events = 0;
	}

	shutdown_clean:
	if(sockfd > 0) close(sockfd);
	for(int i = 0; i < SLOTLIMIT; i++){ if(conns[i] != NULL) free(conns[i]); }
	return 0;

	shutdown_error:
	if(sockfd > 0) close(sockfd);
	return 1;
}


static int write_connection(struct connection *conn)
{
	// Build a filename for the file we're writing based
	// on the time and the source IP address
	char timebuff[20];
	struct tm* timeptr = localtime(&conn->tv.tv_sec);
	strftime(&timebuff[0], 20, "%Y%m%d%H%M%S", timeptr);

	char fnbuff[50];
	snprintf(&fnbuff[0], 50, "%s.%ld-%s.dat",
		timebuff, (long)conn->tv.tv_usec,
		inet_ntoa(conn->addr)
	);

	int ofd;
	if((ofd = open(fnbuff, O_WRONLY | O_CREAT, 0644)) == -1){
		fprintf(stderr, "%s: cannot open - %s\n",
			exec_name, strerror(errno)
		);
		return 1;
	}

	if(write(ofd, conn->buff, conn->len) == -1){
		fprintf(stderr, "%s: file write error - %s\n",
			exec_name, strerror(errno)
		);
		close(ofd);
		return 1;
	}

	close(ofd);
	return 0;
}
